import copy
from typing import Any
from typing import Dict
from typing import List


def update_role_definitions_payload(
    hub, existing_payload: Dict[str, Any], new_values: Dict[str, Any]
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
     PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
         this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
         do an update.

    Returns:
        A result dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages tuple.
    """
    result = {"result": True, "ret": None, "comment": []}
    need_update = False
    new_payload = copy.deepcopy(existing_payload)
    existing_properties = existing_payload["properties"]
    if (new_values.get("role_definition_name") is not None) and (
        new_values["role_definition_name"] != existing_properties.get("roleName")
    ):
        new_payload["properties"]["roleName"] = new_values.get("role_definition_name")
        need_update = True
    if (new_values.get("description") is not None) and (
        new_values["description"] != existing_properties.get("description")
    ):
        new_payload["properties"]["description"] = new_values.get("description")
        need_update = True
    if (new_values.get("permissions") is not None) and (
        compare_permissions(
            new_values["permissions"], existing_properties.get("permissions"), "actions"
        )
    ):
        new_payload["properties"]["permissions"] = new_values.get("permissions")
        need_update = True
    if (new_values.get("assignable_scopes") is not None) and (
        set(new_values["assignable_scopes"])
        != set(existing_properties.get("assignableScopes"))
    ):
        new_payload["properties"]["assignableScopes"] = new_values["assignable_scopes"]
        need_update = True
    if need_update:
        result["ret"] = new_payload
    return result


def convert_raw_role_definitions_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    role_definition_id: str,
    scope: str,
    resource_id: str,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        role_definition_id: Azure role definition resource id.
        scope: scope of the resource.
        resource_id: Role definition resource id on Azure.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "role_definition_id": role_definition_id,
        "scope": scope,
    }
    properties = resource.get("properties")
    if properties:
        properties_parameters = {
            "roleName": "role_definition_name",
            "permissions": "permissions",
            "assignableScopes": "assignable_scopes",
        }
        for parameter_raw, parameter_present in properties_parameters.items():
            if parameter_raw in properties:
                resource_translated[parameter_present] = properties[parameter_raw]
        if properties.get("description") is not None:
            resource_translated["description"] = properties.get("description")
    return resource_translated


def convert_present_to_raw_role_definitions(
    hub,
    scope: str,
    role_definition_name: str,
    permissions: List[str],
    description: str = None,
    assignable_scopes: List[str] = None,
) -> Dict[str, Any]:
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        scope: The scope of the role definition.
        role_definition_name: The role definition name used in the role definition.
        permissions: The permissions of the role definitions.
        description: The description of the role definitions.
        assignable_scopes: The assignable scopes of the role definitions.

    Returns:
        A dict in the format of an Azure PUT operation payload.
    """
    payload = {
        "properties": {
            "roleName": role_definition_name,
            "permissions": permissions,
        }
    }
    if description is not None:
        payload["properties"]["description"] = description
    if assignable_scopes is not None:
        payload["properties"]["assignableScopes"] = assignable_scopes
    else:
        payload["properties"]["assignableScopes"] = [scope]
    return payload


def compare_permissions(
    new_values: List[Dict[str, any]], resource: List[Dict[str, any]], sorting_key: str
):
    """
    Compares role definition permissions to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        new_values: (List[Dict[str, any]]) Present value which will be given as input.
        resource: (List[Dict[str, any]]) Raw resource response which needs to be compared with new_values.
        sorting_key: (str) Primary/Unique key name within each dictionary , which will be used to sort dictionary
         objects with given list before comparing.

    Returns:
        A boolean value, True if there is any difference between List[Dict] arguments else returns False.
    """
    new_values = sorted(new_values, key=lambda x: x[sorting_key])
    resource_list = sorted(resource, key=lambda x: x[sorting_key])
    for i in range(len(resource_list)):
        for key, value in resource_list[i].items():
            if new_values[i].get(key) is None or new_values[i][key] != value:
                return True
    return False


def get_resource_id(hub, resource: Dict[str, any]):
    """
    Returns the resource_id based on the role_definition type if built in role_definition removing the subscription scope from resource_id
    Args:
        resource: (list[dict[str, any]]) Raw resource response which needs to be compared with new_values.

    Returns:
        A string value of resource_id
    """
    resource_id = resource.get("id")
    if resource.get("properties"):
        if (
            resource["properties"].get("type") == "BuiltInRole"
            and resource_id.find("/providers") != -1
        ):
            resource_id = resource_id[resource_id.index("/providers") :]
            hub.log.debug(
                f"Role Definition resource id is being rewritten from {resource['id']} to {resource_id}."
            )
    return resource_id



"""Exec module for managing Virtual Network Peerings."""

__func_alias__ = {"list_": "list"}

from typing import Dict
from collections import OrderedDict


async def get(hub, ctx, resource_id: str, name: str = None, raw: bool = False) -> Dict:
    """Gets Virtual Network Peerings from azure account.

    Args:
        resource_id(str):
            The resource id of the Virtual network.
        name(str, Optional):
            The name of the resource
        raw(bool, Optional):
            Returns raw response if True. Defaults to False


    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec azure.network.virtual_network_peerings.get resource_id="value" raw="False"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path:  azure.network.virtual_network_peerings.get
                - kwargs:
                    resource_id: "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/virtualNetworks/{virtual_network_name}/virtualNetworkPeerings/{virtual_network_peering_name}"
                    raw: "False"
    """
    result = dict(comment=[], ret=None, result=True)
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-03-01",
        success_codes=[200],
    )
    if not response_get["result"]:
        if response_get["status"] != 404:
            result["result"] = False
        result["comment"].extend(
            hub.tool.azure.result_utils.extract_error_comments(response_get)
        )
        return result

    if raw:
        result["ret"] = response_get["ret"]
    else:
        uri_parameters = OrderedDict(
            {
                "subscriptions": "subscriptionId",
                "resourceGroups": "resourceGroupName",
                "virtualNetworks": "virtualNetworkName",
                "virtualNetworkPeerings": "virtualNetworkPeeringName",
            }
        )
        uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
            resource_id, uri_parameters
        )
        result[
            "ret"
        ] = hub.tool.azure.network.virtual_network_peerings.convert_raw_to_present_state(
            {**uri_parameter_values, **response_get["ret"]}
        )

        result["ret"]["name"] = resource_id
        result["ret"]["resource_id"] = resource_id
    return result


async def list_(
    hub,
    ctx,
) -> Dict:
    """Lists all Virtual Network Peerings.

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id

        .. code-block:: bash

            idem exec azure.network.virtual_network_peerings.list

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: azure.network.virtual_network_peerings.list

    """
    result = dict(comment=[], ret=[], result=True)
    subscription_id = ctx.acct.subscription_id

    virtual_networks_list = await hub.exec.azure.network.virtual_networks.list(ctx)
    if not virtual_networks_list["result"]:
        result["result"] = False
        result["comment"].append(virtual_networks_list["comment"])
        return result

    for vn in virtual_networks_list["ret"]:
        vn_name = vn["virtual_network_name"]
        resource_group_name = vn["resource_group_name"]
        uri_parameters = OrderedDict(
            {
                "subscriptions": "subscriptionId",
                "resourceGroups": "resourceGroupName",
                "virtualNetworks": "virtualNetworkName",
                "virtualNetworkPeerings": "virtualNetworkPeeringName",
            }
        )
        async for page_result in hub.tool.azure.request.paginate(
            ctx,
            url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/virtualNetworks/{vn_name}/virtualNetworkPeerings?api-version=2021-03-01",
            success_codes=[200, 404],
        ):
            resource_list = page_result.get("value", None)
            if resource_list:
                for resource_raw_state in resource_list:
                    resource_id = resource_raw_state["id"]
                    uri_parameter_values = (
                        hub.tool.azure.uri.get_parameter_value_in_dict(
                            resource_id, uri_parameters
                        )
                    )
                    resource_present_state = hub.tool.azure.network.virtual_network_peerings.convert_raw_to_present_state(
                        {**uri_parameter_values, **resource_raw_state}
                    )
                    resource_present_state["resource_id"] = resource_id
                    resource_present_state["name"] = resource_id

                    result["ret"].append(resource_present_state)

    return result



from typing import Any
from typing import Dict


def convert_raw_attach_subscription_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    management_group_id: str,
    subscription_id: str,
    resource_id: str,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing attached management group subscription state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the management group subscription.
        management_group_id: Management Group ID.
        subscription_id: Subscription ID which needs to be attached to MG.
        resource_id: Azure Management group subscription association ID.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "management_group_id": management_group_id,
        "subscription_id": subscription_id,
    }
    return resource_translated


def convert_present_to_raw_subscription(
    hub,
    billing_scope: str = None,
    display_name: str = None,
    workload: str = None,
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        billing_scope: billing scope associated with billing account id and enrollment account id to create subscription
        display_name: Display name of subscription
        workload: Workload type for subscription which can be Production/DevTest

    Returns:
        A dict in the format of an Azure PUT operation payload.
    """
    payload = {
        "properties": {
            "billingScope": billing_scope,
            "DisplayName": display_name,
            "Workload": workload,
        }
    }

    if billing_scope is not None:
        payload["properties"]["billingScope"] = billing_scope
    if display_name is not None:
        payload["properties"]["DisplayName"] = display_name
    if workload is not None:
        payload["properties"]["Workload"] = workload
    return payload


def convert_raw_subscription_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    resource_id: str,
    alias: str = None,
    display_name: str = None,
    subscription_id: str = None,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        alias: Alias for subscription
        display_name: Display name of subscription
        subscription_id: Unique subscription id
        resource_id: Azure subscription resource id.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "alias": alias,
        "display_name": display_name,
        "subscription_id": subscription_id,
    }

    if display_name is None:
        resource_translated["display_name"] = resource.get("displayName")
    if subscription_id is None:
        properties = resource.get("properties")
        if properties:
            resource_translated["subscription_id"] = properties.get("subscriptionId")
        else:
            resource_translated["subscription_id"] = resource.get("subscriptionId")
    return resource_translated



import asyncio
from dataclasses import dataclass
from typing import Any
from typing import Dict
from typing import Optional

OPERATION_TRACKERS = {}


@dataclass
class OperationResult:
    finished: bool
    succeeded: bool
    error_message: str = None
    status: int = None
    operation_url: str = None
    resource: Dict[str, Any] = None


# result is True only if operation finished successfully
# if it is still running -> result is False and there is rerun_data containing information passed on input
# it long-running operation finished with an error -> result is False and there is rerun_data == {"has_error": True}
# if get_resource is True and operation has succeeded, get the resource by the resource_url in rerun_data
# resource is returned in result["resource"]
async def handle_operation(
    hub, ctx, rerun_data, get_resource: bool = True
) -> Dict[str, Any]:
    """
    Processes the rerun_data of a long-running asynchronous operation
    Fetches the current operation state and does error handling if needed
    Returns either the end result or the rerun_data for subsequent iterations
    :param ctx: Context to be used for Azure API calls
    :param rerun_data: Object containing any information needed for processing
        Should include the following keys:
        - operation_id - unique identifier for operation within the idem process
        - operation_headers - headers returned by Azure API needed for fetching the operation information
        - resource_url - only required if get_resource is true - URL of the targeted resource to be fetched upon operation success
        Any other keys get ignored and passed through transparently in resulting rerun_data if operation is still running
    :param get_resource: True if caller wants the method to fetch the targeted resource upon success. Resource returned in result["resource]
    """
    result = {
        "comment": [],
        "result": True,
        "rerun_data": None,
        "resource": None,
    }

    operation_id = rerun_data.get("operation_id")

    if operation_id is None:
        result["result"] = False
        result["comment"].append("Operation ID is not supplied")
        result["rerun_data"] = {"has_error": True}
        return result

    try:
        operation_result: OperationResult = (
            await hub.tool.azure.operation_utils.check_operation_result(
                ctx,
                operation_id=operation_id,
                operation_headers=rerun_data.get("operation_headers"),
            )
        )
    except Exception as e:
        result["result"] = False
        result["comment"].append(
            "An error occurred while processing the long-running operation"
        )
        result["comment"].append(str(e))
        result["rerun_data"] = {"has_error": True}
        return result

    if not operation_result.finished:
        # Operation still hasn't finished -> reconcile.
        result["rerun_data"] = rerun_data
        result["result"] = False
        return result

    if not operation_result.succeeded:
        result["rerun_data"] = {"has_error": True}
        result["comment"].append(
            f"Long running operation {operation_result.operation_url} failed"
        )
        if operation_result.status:
            result["comment"].append(f"Status code: {operation_result.status}")
        if operation_result.error_message:
            result["comment"].append(operation_result.error_message)
        result["result"] = False
        return result

    if not get_resource:
        # no further processing needed
        return result

    resource_url = rerun_data.get("resource_url")
    if not resource_url:
        raise RuntimeError("Resource URL not provided!")

    get_result = await hub.exec.request.json.get(
        ctx,
        url=resource_url,
        success_codes=[200],
    )

    if get_result.get("result"):
        result["resource"] = get_result.get("ret") or None
    elif get_result.get("status") == 404:
        result["resource"] = None
    else:
        result["result"] = False
        result["rerun_data"] = {"has_error": True}
        result["comment"].append(
            f"Long running operation {operation_result.operation_url} succeeded, but get request for {resource_url} failed"
        )
        if get_result.get("status"):
            result["comment"].append(f"Status code: {operation_result.status}")
        if get_result.get("comment"):
            result["comment"].append(get_result["comment"])
        if get_result.get("ret") and "error" in get_result.get("ret"):
            result["comment"].append(str(get_result["ret"]["error"]))

    return result


async def await_operation(hub, ctx, operation_headers, resource_url) -> Dict[str, Any]:
    """
    Awaits a long-running asynchronous Azure operation to complete. Performs an additional GET request when operation completes to fetch the resulting Azure resource.
    :param operation_headers: Headers from initial HTTP response from Azure API call which initiated the operation
    :param resource_url: Azure API URL, used to GET the resource produced by the operation
    """
    poller = OperationPoller(hub, ctx, operation_headers, resource_url)
    return await poller.await_operation()


async def check_operation_result(
    hub, ctx, operation_id, operation_headers
) -> OperationResult:
    """
    Returns the current status of a long-running asynchronous Azure operation.
    :param operation_id: Client-supplied unique operation ID
    :param operation_headers: Headers from initial HTTP response from Azure API call which initiated the operation
    """
    poller = OPERATION_TRACKERS.get(operation_id)
    if not poller:
        poller = OperationPoller(hub, ctx, operation_headers)
        OPERATION_TRACKERS[operation_id] = poller
    try:
        operation_result = await poller.check_operation()
        if operation_result.finished:
            del OPERATION_TRACKERS[operation_id]
        return operation_result
    except Exception as e:
        op_url = poller.get_operation_url()
        del OPERATION_TRACKERS[operation_id]
        return OperationResult(
            finished=True, succeeded=False, error_message=str(e), operation_url=op_url
        )


class OperationFailed(Exception):
    pass


class OperationStatus:
    """Operation status class.

    Operation status is used to indicate the status of an operation. It can be one of the following values: Succeeded, Failed, Canceled, Running.
    """

    SUCCEEDED = "Succeeded"
    FAILED = "Failed"
    CANCELED = "Canceled"
    RUNNING = "Running"

    FINISHED_STATES = {SUCCEEDED.lower(), FAILED.lower(), CANCELED.lower()}
    FAILED_STATES = {FAILED.lower(), CANCELED.lower()}
    SUCCEEDED_STATES = {SUCCEEDED.lower()}

    @staticmethod
    def finished(status):
        return str(status).lower() in OperationStatus.FINISHED_STATES

    @staticmethod
    def failed(status):
        return str(status).lower() in OperationStatus.FAILED_STATES

    @staticmethod
    def succeeded(status):
        return str(status).lower() in OperationStatus.SUCCEEDED_STATES


class OperationPoller:
    def __init__(self, hub, ctx, operation_headers, resource_url: str = None):
        self._hub = hub
        self._ctx = ctx
        self._operation_headers = operation_headers
        self._resource_url = resource_url
        self._status_url = self._get_status_url()
        self._status = None
        self._error = None

    async def await_operation(self) -> Dict[str, Any]:
        return await self._poll()

    def get_operation_url(self) -> str:
        return self._status_url

    async def check_operation(self) -> OperationResult:
        if not self.operation_finished():
            await self._update_status()
        return OperationResult(
            self.operation_finished(),
            self._operation_succeeded(),
            self._error,
            self._status,
            operation_url=self._status_url,
        )

    def _get_status_url(self) -> Optional[str]:
        headers: Dict[str, Any] = self._operation_headers

        if "Azure-AsyncOperation" in headers:
            return headers["Azure-AsyncOperation"]
        elif "Location" in headers:
            return headers["Location"]
        else:
            return None

    def _get_retry_after(self) -> int:
        headers = self._operation_headers

        retry_after = headers.get("Retry-After")
        try:
            delay = int(retry_after)
        except:
            delay = 1

        return delay if delay > 1 else 1

    def operation_finished(self) -> bool:
        return OperationStatus.finished(self._status)

    def _operation_succeeded(self) -> bool:
        return OperationStatus.succeeded(self._status)

    async def _update_status(self):
        get_response = await self._hub.exec.request.json.get(
            self._ctx,
            url=self._status_url,
            success_codes=[200],
        )
        if "status" in get_response["ret"]:
            self._status = get_response["ret"]["status"].lower()
        else:
            self._status = None
        if "error" in get_response["ret"]:
            self._error = get_response["ret"]["error"]
        else:
            self._error = None

    async def _delay(self):
        await asyncio.sleep(self._get_retry_after())

    async def _poll(self):
        if not self.operation_finished():
            await self._update_status()
        while not self.operation_finished():
            await self._delay()
            await self._update_status()

        if not self._operation_succeeded():
            raise OperationFailed(
                f"Operation failed or has been canceled: {self._error}"
            )
        else:
            get_response = await self._hub.exec.request.json.get(
                self._ctx,
                url=self._resource_url,
                success_codes=[200],
            )
            return get_response



import re


try:
    import bs4
    import requests

    HAS_LIBS = (True,)
except ImportError as e:
    HAS_LIBS = False, str(e)


def __virtual__(hub):
    return HAS_LIBS


HEADERS = {
    "user-agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10.9; rv:32.0) Gecko/20100101 Firefox/32.0"
}


def parse(hub, url: str, desc: str):
    """
    Parse a url like this: https://docs.microsoft.com/en-us/rest/api/power-bi/available-features
    Check all the "Operations" for Get, List, Delete, and Create
    If we have all of those then create a plugin with "present", "absent", and "describe" functions
    """
    plugin = {}
    api_plugin = {}
    relative_links = set()

    # Define a helper function to determine resources available given a URL
    def resource_(url, page):
        """
        Find the primary API group given a URL.

        Since the URL has the primary candidate for a valid API, check the page itself
        and determine if it's modified. If so, return the correct format.
        :return:
        """
        potential = None
        tag = None
        op_groups = []
        candidate = url.strip("/").split("/")[-1]

        # Find a secondary candidate and cache the operation tag objects while we have them
        if page.find("h2", text="REST operation groups"):
            op_groups = page.find("h2", text="REST operation groups").parent.find_all(
                "td"
            )
            tag = op_groups[0]
        elif page.find("h2", {"id": "rest-operation-groups"}):
            op_groups = page.find(
                "h2", {"id": "rest-operation-groups"}
            ).parent.find_all("td")
            tag = op_groups[0]
        else:
            pass

        try:
            name_dict = tag.contents[0].attrs
        except Exception:
            hub.log.debug(f"We caught a general error for {url}")
        else:
            potential = name_dict.get("href")

        # Determine if modifications are needed and send them back
        temp = re.sub(r"\W+", "", candidate)
        if not potential or temp in potential:
            base_url = url.split(candidate)[0]
            url = base_url + temp
            candidate = temp
        return candidate, op_groups, url

    with requests.get(url) as response:
        html = response.text

    page = bs4.BeautifulSoup(html, "lxml")
    api, operation_groups, url = resource_(url, page)

    for tag in operation_groups:
        try:
            temp = tag.contents[0].attrs.get("href").split(api)[-1].strip("/")
        except (
            AttributeError,
            IndexError,
        ):
            # Exception thrown because parsing a tag doesn't contain a valid resource so skip it
            hub.log.debug(
                f"We caught an AttributeError or IndexError for {url}\nThis is likely because a resource was not valid so pass."
            )
        else:
            resource = temp
            if not resource == "relative-path" and resource is not None:
                rel_link = f"{url}/{resource}"
                relative_links.add(rel_link)

    # Determine operations on a per resource basis
    operation_groups = []
    for link in relative_links:
        # Create a local list of acceptable article declarations
        resource = link.split("/")[-1]
        allowables = [
            f"{resource}/create",
            f"{resource}/create-or-update",
            f"{resource}/update",
            f"{resource}/update-tags",
            f"{resource}/get",
            f"{resource}/delete",
            f"{resource}/list",
            f"{resource}/list-all",
        ]
        with requests.get(link, headers=HEADERS) as resp:
            html = resp.text
        page = bs4.BeautifulSoup(html, "lxml")

        # Create a local cache of links to build return after we find all available
        api_cache = []
        for item in page.find_all("a", {"data-linktype": "relative-path"}):
            resource_ext = item.attrs["href"]
            plugin["resource"] = resource
            if resource_ext in allowables:
                api_cache.append(resource_ext)
        for entry in api_cache:
            entry = f"{url}/{entry}"
            # Build out the return dict
            if "create" in entry:
                plugin["create"] = entry
            elif "update" in entry:
                plugin["update"] = entry
            elif "get" in entry:
                plugin["get"] = entry
            elif "delete" in entry:
                plugin["delete"] = entry
            elif "list-all" in entry:
                plugin["list"] = entry
            elif "list" in entry:
                plugin["list"] = entry
        operation_groups.append(plugin.copy())
    api_plugin[api] = operation_groups
    return api_plugin



def _is_within(parent, o, ignore: set):
    """
    Determine of an object is within a parent object.
    :param parent: The object in which o hopefully exists.
    :param o: The object to find in parent.
    :param ignore: A set of keys to ignore in parent.
    return: True if o is within parent somewhere. False otherwise.
    """
    if not isinstance(parent, type(o)):
        return False
    elif isinstance(o, dict):
        return _is_within_dict(parent, o, ignore)
    else:
        return parent == o


def _is_within_dict(parent, o, ignore: set):
    """
    Determine of an object is within a parent dict object.
    :param parent: The object in which o hopefully exists.
    :param o: The object to find in parent.
    :param ignore: A set of keys to ignore in parent.
    return: True if o is within parent somewhere. False otherwise.
    """
    ret = False
    for k, v in o.items():
        if k in ignore:
            ret = True
            break
        elif k in parent and parent[k] == v:
            ret = True
            break
    return ret


def get_from_list(hub, parent, o):
    """
    Returns the first object found in parent that contains o, None if no object is found.
    :param hub: The redistributed pop central hub. This is required in
    Idem, so while not used, must appear.
    :param parent: An iterable object to search for o.
    :param o: An object the values of which must exist in one of the iterables.

    For example:

    The subset:

    { name: "my_object" }

    exists within (and would be returned)

    [
        { name: "not_my_object", something_else: "not the other thing" }
        { name: "my_object", something_else: "some other thing"},
    ]
    """
    ret = None

    for item in parent:
        if isinstance(item, dict):
            if _is_within(item, o, set()):
                ret = item
                break
        else:
            hub.log.warning(f"Item {str(item)} should be a dict. Skipping this item.")

    return ret



"""State module for managing Virtual Network Peerings."""
from dataclasses import make_dataclass
from typing import Any
from typing import Dict


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    resource_group_name: str = None,
    virtual_network_name: str = None,
    virtual_network_peering_name: str = None,
    sync_remote_address_space: bool = None,
    allow_virtual_network_access: bool = None,
    allow_forwarded_traffic: bool = None,
    allow_gateway_transit: bool = None,
    use_remote_gateways: bool = None,
    remote_virtual_network: make_dataclass(
        "Id",
        [
            ("id", str),
        ],
    ) = None,
    do_not_verify_remote_gateways: bool = None,
    subscription_id: str = None,
) -> Dict[str, Any]:
    r"""Create or update Virtual Network Peerings.

    Args:
        name(str): The identifier for this state.
        resource_group_name(str, Optional): The name of the resource group.
        virtual_network_name(str, Optional): The name of the virtual network.
        virtual_network_peering_name(str, Optional): The name of the peering.
        sync_remote_address_space(bool, Optional): Parameter indicates the intention to sync the peering with the current address space on the remote vNet after it's updated.
        resource_id(str, Optional): Virtual Network resource id on Azure.
        allow_virtual_network_access(bool, Optional):
            Whether the VMs in the local virtual network space would be able to access the VMs in remote virtual network space.
        allow_forwarded_traffic(bool, Optional):
            Whether the forwarded traffic from the VMs in the local virtual network will be allowed/disallowed in remote virtual network.
        allow_gateway_transit(bool, Optional):
            If gateway links can be used in remote virtual networking to link to this virtual network.
        use_remote_gateways(bool, Optional):
            If remote gateways can be used on this virtual network. If the flag is set to true, and allowGatewayTransit on remote peering is also true, virtual network will use gateways of remote virtual network for transit. Only one peering can have this flag set to true. This flag cannot be set if virtual network already has a gateway.
        remote_virtual_network(Dict):

            * id(str):
                The reference to the remote virtual network. The remote virtual network can be in the same or different region (preview). See here to register for the preview and learn more (https://docs.microsoft.com/en-us/azure/virtual-network/virtual-network-create-peering).
        do_not_verify_remote_gateways(bool, Optional):
            If we need to verify the provisioning state of the remote gateway.
        subscription_id(str, Optional): Subscription Unique id.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            my-vnet-peerings:
              azure.network.virtual_network_peerings.present:
                - virtual_network_name: network-1
                - resource_group_name: resource-group-name
                - virtual_network_peering_name: test-peering1
                - remote_virtual_network:
                    id: /subscriptions/sub-id/resourceGroups/resource-group-name/providers/Microsoft.Network/virtualNetworks/network-2
                - use_remote_gateways: False
                - allow_virtual_network_access: False
    """
    path_properties = {
        "subscription_id": subscription_id,
        "resource_group_name": resource_group_name,
        "virtual_network_name": virtual_network_name,
        "virtual_network_peering_name": virtual_network_peering_name,
    }
    query_properties = {"sync_remote_address_space": sync_remote_address_space}
    body_properties = {
        "allow_virtual_network_access": allow_virtual_network_access,
        "allow_forwarded_traffic": allow_forwarded_traffic,
        "allow_gateway_transit": allow_gateway_transit,
        "use_remote_gateways": use_remote_gateways,
        "remote_virtual_network": remote_virtual_network,
        "do_not_verify_remote_gateways": do_not_verify_remote_gateways,
    }
    return await hub.tool.azure.generic.run_present(
        ctx,
        "azure.network.virtual_network_peerings",
        name,
        resource_id,
        path_properties,
        query_properties,
        body_properties,
    )


async def absent(
    hub,
    ctx,
    name: str,
    resource_id: str = None,
    resource_group_name: str = None,
    virtual_network_name: str = None,
    virtual_network_peering_name: str = None,
    subscription_id: str = None,
) -> Dict[str, Any]:
    r"""Delete Virtual Network Peerings.

    Args:
        name(str): The identifier for this state.
        resource_id(str, Optional): Virtual Network resource id on Azure.
        resource_group_name(str): The name of the resource group.
        virtual_network_name(str): The name of the virtual network.
        virtual_network_peering_name(str): The name of the virtual network peering.
        subscription_id(str, Optional): Subscription Unique id.

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: sls

            resource_is_absent:
              azure.network.virtual_network_peerings.absent:
                - resource_group_name: value
                - virtual_network_name: value
                - virtual_network_peering_name: value

    """
    return hub.tool.azure.result_utils.absent_implemented_through_wrapper_result(
        "azure.network.virtual_network_peerings", name
    )


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all Virtual Network Peerings under the same subscription

    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe azure.network.virtual_network_peerings
    """
    result = {}
    ret_list = await hub.exec.azure.network.virtual_network_peerings.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(
            f"Could not describe network virtual_network_peerings {ret_list['comment']}"
        )
        return result

    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.network.virtual_network_peerings.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result


def is_pending(hub, ret: dict, state: str = None, **pending_kwargs) -> bool:
    """Default implemented for each module."""
    return hub.tool.azure.resource_utils.is_pending(
        ret=ret, state=state, **pending_kwargs
    )



from typing import Any
from typing import Dict

from dict_tools.data import NamespaceDict


async def gather(hub, profiles) -> Dict[str, Any]:
    """
    Get a new access token based on Azure client_id, client_secret and tenant_id

    Example:
    .. code-block:: yaml

        azure:
          profile_name:
            client_id: Azure client id "aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa"
            secret: Azure client secret "xxxxxxxxxxxxxxxxxxxxxxxxxxxxxx"
            subscription_id: Azure subscription id "bbbbbbbb-bbbb-bbbb-bbbb-bbbbbbbbbbbb"
            tenant: Azure tenant id "cccccccc-cccc-cccc-cccc-cccccccccccc"
            api_version: 2015-07-01
    """
    sub_profiles = {}
    for (
        profile,
        ctx,
    ) in profiles.get("azure", {}).items():
        temp_ctx = NamespaceDict(acct={})
        tenant = ctx["tenant"]
        client_id = ctx["client_id"]
        secret = ctx["secret"]
        subscription_id = ctx["subscription_id"]

        login_endpoint_url = (
            ctx.get("login_endpoint_url")
            or ctx.get("endpoint_url")
            or hub.exec.azure.LOGIN_URL
        )
        resource_endpoint_url = ctx.get("endpoint_url") or hub.exec.azure.URL

        result = await hub.exec.request.json.post(
            temp_ctx,
            url=f"{login_endpoint_url}/{tenant}/oauth2/token",
            headers={
                "content-type": "application/x-www-form-urlencoded",
            },
            data={
                "grant_type": "client_credentials",
                "client_id": client_id,
                "client_secret": secret,
                "resource": resource_endpoint_url,
            },
        )

        if not result["status"] or not result["result"]:
            status_comment = result.get("comment", "")
            error_comment = result.get("ret", "")
            error = f"Unable to authenticate with tenant '{tenant}' and client id '{client_id}': {status_comment} {error_comment}"
            hub.log.error(error)
            raise ConnectionError(error)

        ret = result.get("ret")
        if not ret or not isinstance(ret, dict):
            error = f"result['ret'] must be of type dict. Actual: ret type: {type(ret)}, ret value: {ret}"
            hub.log.error(error)
            raise TypeError(error)

        access_token = result["ret"].get("access_token")
        sub_profiles[profile] = dict(
            subscription_id=subscription_id,
            endpoint_url=resource_endpoint_url,
            login_endpoint_url=login_endpoint_url,
            headers={"Authorization": f"Bearer {access_token}"},
        )
    return sub_profiles



"""State module for managing Resource Group."""
import copy
from typing import Any
from typing import Dict


__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    location: str,
    subscription_id: str = None,
    resource_id: str = None,
    tags: Dict = None,
) -> dict:
    r"""Create or update Resource Groups.

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group to create or update. Can include alphanumeric, underscore, parentheses, hyphen, period (except at end), and Unicode characters that match the allowed characters.Regex pattern: ^[-\w\._\(\)]+$.
        location(str): Resource location.
        subscription_id(str, Optional): Subscription Unique id.
        resource_id(str, Optional): Resource Group id on Azure.
        tags(dict, Optional): Resource tags.

    Returns:
        dict

    Examples:
        .. code-block:: sls

            resource_is_present:
              azure.resource_management.resource_groups.present:
                - name: value
                - resource_group_name: value
                - subscription_id: value
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    if resource_id is None:
        resource_id = (
            f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}"
        )
    response_get = await hub.exec.azure.resource_management.resource_groups.get(
        ctx, resource_id=resource_id, raw=True
    )
    if response_get["result"]:
        if response_get["ret"] is None:
            if ctx.get("test", False):
                # Return a proposed state by Idem state --test
                result[
                    "new_state"
                ] = hub.tool.azure.test_state_utils.generate_test_state(
                    enforced_state={},
                    desired_state={
                        "name": name,
                        "resource_group_name": resource_group_name,
                        "subscription_id": subscription_id,
                        "tags": tags,
                        "location": location,
                        "resource_id": resource_id,
                    },
                )
                result["comment"].append(
                    f"Would create azure.resource_management.resource_groups '{name}'"
                )
                return result

            else:
                # PUT operation to create a resource
                payload = hub.tool.azure.resource_management.resource_groups.convert_present_to_raw_resource_group(
                    location=location,
                    tags=tags,
                )
                response_put = await hub.exec.request.json.put(
                    ctx,
                    url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-04-01",
                    success_codes=[200, 201],
                    json=payload,
                )

                if not response_put["result"]:
                    hub.log.debug(
                        f"Could not create Resource Groups {response_put['comment']} {response_put['ret']}"
                    )
                    result["comment"].extend(
                        hub.tool.azure.result_utils.extract_error_comments(response_put)
                    )
                    result["result"] = False
                    return result

            result[
                "new_state"
            ] = hub.tool.azure.resource_management.resource_groups.convert_raw_resource_group_to_present(
                resource=response_put["ret"],
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                resource_id=resource_id,
                subscription_id=subscription_id,
            )
            result["comment"].append(
                f"Created azure.resource_management.resource_groups '{name}'"
            )
            return result
        else:
            existing_resource = response_get["ret"]
            result[
                "old_state"
            ] = hub.tool.azure.resource_management.resource_groups.convert_raw_resource_group_to_present(
                resource=existing_resource,
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                resource_id=resource_id,
                subscription_id=subscription_id,
            )
            # Generate a new PUT operation payload with new values
            new_payload = hub.tool.azure.resource_management.resource_groups.update_resource_groups_payload(
                existing_resource,
                {"tags": tags},
            )
            if ctx.get("test", False):
                if new_payload["ret"] is None:
                    result["new_state"] = copy.deepcopy(result["old_state"])
                    result["comment"].append(
                        f"azure.resource_management.resource_groups '{name}' has no property need to be updated."
                    )
                else:
                    result[
                        "new_state"
                    ] = hub.tool.azure.resource_management.resource_groups.convert_raw_resource_group_to_present(
                        resource=new_payload["ret"],
                        idem_resource_name=name,
                        resource_group_name=resource_group_name,
                        resource_id=resource_id,
                        subscription_id=subscription_id,
                    )
                    result["comment"].append(
                        f"Would update azure.resource_management.resource_groups '{name}'"
                    )
                return result
            # PUT operation to update a resource
            if new_payload["ret"] is None:
                result["new_state"] = copy.deepcopy(result["old_state"])
                result["comment"].append(
                    f"azure.resource_management.resource_groups '{name}' has no property need to be updated."
                )
                return result
            result["comment"].extend(new_payload["comment"])
            response_put = await hub.exec.request.json.put(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-04-01",
                success_codes=[200, 201],
                json=new_payload["ret"],
            )

            if not response_put["result"]:
                hub.log.debug(
                    f"Could not update azure.resource_management.resource_groups {response_put['comment']} {response_put['ret']}"
                )
                result["result"] = False
                result["comment"].extend(
                    hub.tool.azure.result_utils.extract_error_comments(response_put)
                )
                return result

            result[
                "new_state"
            ] = hub.tool.azure.resource_management.resource_groups.convert_raw_resource_group_to_present(
                resource=response_put["ret"],
                idem_resource_name=name,
                resource_group_name=resource_group_name,
                resource_id=resource_id,
                subscription_id=subscription_id,
            )
            result["comment"].append(
                f"Updated azure.resource_management.resource_groups '{name}'"
            )
            return result
    else:
        hub.log.debug(
            f"Could not get azure.resource_management.resource_groups {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"].extend(
            hub.tool.azure.result_utils.extract_error_comments(response_get)
        )
        return result


async def absent(
    hub, ctx, name: str, resource_group_name: str, subscription_id: str = None
) -> dict:
    r"""Delete Resource Groups.

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group to delete. The name is case insensitive.Regex pattern: ^[-\w\._\(\)]+$.
        subscription_id(str, Optional): Subscription Unique id.

    Returns:
        dict

    Examples:
        .. code-block:: sls

            resource_is_absent:
              azure.resource_management.resource_groups.absent:
                - name: value
                - resource_group_name: value
                - subscription_id: value
    """
    result = {
        "name": name,
        "result": True,
        "old_state": None,
        "new_state": None,
        "comment": [],
    }
    if subscription_id is None:
        subscription_id = ctx.acct.subscription_id
    resource_id = (
        f"/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}"
    )
    response_get = await hub.exec.azure.resource_management.resource_groups.get(
        ctx, resource_id=resource_id
    )
    if response_get["result"]:
        if response_get["ret"]:
            result["old_state"] = response_get["ret"]
            result["old_state"]["name"] = name
            if ctx.get("test", False):
                result["comment"].append(
                    f"Would delete azure.resource_management.resource_groups '{name}'"
                )
                return result
            response_delete = await hub.exec.request.raw.delete(
                ctx,
                url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-04-01",
                success_codes=[200, 202],
            )

            if not response_delete["result"]:
                hub.log.debug(
                    f"Could not delete azure.resource_management.resource_groups {response_delete['comment']} {response_delete['ret']}"
                )
                result["result"] = False
                result["comment"].extend(
                    hub.tool.azure.result_utils.extract_error_comments(response_delete)
                )
                return result

            result["comment"].append(
                f"Deleted azure.resource_management.resource_groups '{name}'"
            )
            return result
        else:
            # If Azure returns 'Not Found' error, it means the resource has been absent.
            result["comment"].append(
                f"azure.resource_management.resource_groups '{name}' already absent"
            )
            return result
    else:
        hub.log.debug(
            f"Could not get azure.resource_management.resource_groups '{name}' {response_get['comment']} {response_get['ret']}"
        )
        result["result"] = False
        result["comment"].extend(
            hub.tool.azure.result_utils.extract_error_comments(response_get)
        )
    return result


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""Describe the resource in a way that can be recreated/managed with the corresponding "present" function.

    Lists all Resource Groups under the same subscription.


    Returns:
        Dict[str, Any]

    Examples:
        .. code-block:: bash

            $ idem describe azure.resource_management.resource_groups
    """
    result = {}
    ret_list = await hub.exec.azure.resource_management.resource_groups.list(ctx)
    if not ret_list["ret"]:
        hub.log.debug(f"Could not describe resource_groups {ret_list['comment']}")
        return result

    for resource in ret_list["ret"]:
        resource_id = resource["resource_id"]
        result[resource_id] = {
            "azure.resource_management.resource_groups.present": [
                {parameter_key: parameter_value}
                for parameter_key, parameter_value in resource.items()
            ]
        }

    return result



import copy
from typing import Any
from typing import Dict
from typing import List


def convert_raw_virtual_network_to_present(
    hub,
    resource: Dict,
    idem_resource_name: str,
    resource_group_name: str,
    virtual_network_name: str,
    resource_id: str,
    subscription_id: str,
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        hub: The redistributed pop central hub.
        resource: An existing resource state from Azure. This is usually a GET operation response.
        idem_resource_name: The Idem name of the resource.
        resource_group_name: Azure Resource Group name.
        virtual_network_name: Azure Virtual Network resource name.
        resource_id: Azure Virtual Network resource id.
        subscription_id: The Microsoft Azure subscription ID.

    Returns:
        A dict that contains the parameters that match the present function's input format.
    """
    resource_translated = {
        "name": idem_resource_name,
        "resource_id": resource_id,
        "resource_group_name": resource_group_name,
        "virtual_network_name": virtual_network_name,
        "location": resource["location"],
        "subscription_id": subscription_id,
    }
    if "tags" in resource:
        resource_translated["tags"] = resource["tags"]
    if "extendedLocation" in resource:
        resource_translated["extended_location"] = resource["extendedLocation"]
    properties = resource.get("properties")
    if properties:
        properties_parameters = {
            "flowTimeoutInMinutes": "flow_timeout_in_minutes",
            "ddosProtectionPlan": "ddos_protection_plan",
            "enableDdosProtection": "enable_ddos_protection",
            "enableVmProtection": "enable_vm_protection",
        }
        for parameter_raw, parameter_present in properties_parameters.items():
            if parameter_raw in properties:
                resource_translated[parameter_present] = properties.get(parameter_raw)
        if (
            properties.get("addressSpace") is not None
            and properties.get("addressSpace", {}).get("addressPrefixes") is not None
        ):
            resource_translated["address_space"] = properties["addressSpace"][
                "addressPrefixes"
            ]
        if properties.get("subnets") is not None:
            existing_subnets_required_payload = convert_raw_to_present_vn_subnet(
                subnets=properties["subnets"]
            )
            resource_translated["subnets"] = existing_subnets_required_payload
        if (
            properties.get("dhcpOptions") is not None
            and properties.get("dhcpOptions", {}).get("dnsServers") is not None
        ):
            resource_translated["dhcp_options"] = {
                "dns_servers": properties["dhcpOptions"]["dnsServers"]
            }
        if (
            properties.get("bgpCommunities") is not None
            and properties.get("bgpCommunities", {}).get("virtualNetworkCommunity")
            is not None
        ):
            resource_translated["bgp_communities"] = {
                "virtual_network_community": properties["bgpCommunities"][
                    "virtualNetworkCommunity"
                ]
            }
    return resource_translated


def convert_present_to_raw_virtual_network(
    hub,
    subscription_id: str,
    address_space: List,
    location: str,
    extended_location: Dict = None,
    bgp_communities: Dict = None,
    flow_timeout_in_minutes: int = None,
    subnets: List = None,
    ddos_protection_plan: Dict = None,
    enable_ddos_protection: bool = None,
    enable_vm_protection: bool = None,
    dhcp_options: Dict = None,
    tags: Dict = None,
    **kwargs
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        hub: The redistributed pop central hub.
        subscription_id: subscription id
        address_space: An array of IP address ranges that can be used by subnets of the virtual network.
        location: Resource location. Update this field will result in resource re-creation.
        extended_location: The extended location of the virtual network.
        bgp_communities: Bgp Communities sent over ExpressRoute with each route corresponding to a prefix in this VNET.
        flow_timeout_in_minutes: The FlowTimeout value (in minutes) for the Virtual Network
        subnets(list[dict], Optional): List of Subnet in a virtual network resource.Each Subnet will have fields
          name(str, required), address_space(str, required) , security_group_id(str, optional) and service_endpoints(list, optional)
        ddos_protection_plan(Dict, Optional):
            id(str): The DDoS protection plan associated with the virtual network.
        enable_ddos_protection(bool, Optional): Indicates if DDoS protection is enabled for all the protected resources in the virtual network. It requires a DDoS protection plan associated with the resource.
        enable_vm_protection(bool, Optional): Indicates if VM protection is enabled for all the subnets in the virtual network.
        dhcp_options(Dict, Optional): DhcpOptions contains an array of DNS servers available to VMs deployed in the virtual network. Standard DHCP option for a subnet overrides VNET DHCP options.
        tags: Resource tags.

    Returns:
        A dict in the format of an Azure PUT operation payload.
    """
    payload = {
        "location": location,
        "properties": {},
    }
    if tags is not None:
        payload["tags"] = tags
    if extended_location is not None:
        payload["extendedLocation"] = extended_location
    if address_space is not None:
        payload["properties"]["addressSpace"] = {"addressPrefixes": address_space}
    if (
        bgp_communities is not None
        and bgp_communities.get("virtual_network_community") is not None
    ):
        payload["properties"]["bgpCommunities"] = {
            "virtualNetworkCommunity": bgp_communities["virtual_network_community"]
        }
    if flow_timeout_in_minutes is not None:
        payload["properties"]["flowTimeoutInMinutes"] = flow_timeout_in_minutes
    if ddos_protection_plan is not None and ddos_protection_plan.get("id") is not None:
        payload["properties"]["ddosProtectionPlan"] = ddos_protection_plan
    if enable_ddos_protection is not None:
        payload["properties"]["enableDdosProtection"] = enable_ddos_protection
    if enable_vm_protection is not None:
        payload["properties"]["enableVmProtection"] = enable_vm_protection
    if subnets is not None:
        subnets_payload = convert_present_to_raw_vn_subnet(
            subscription_id=subscription_id, subnets=subnets
        )
        payload["properties"]["subnets"] = subnets_payload
    if dhcp_options is not None and dhcp_options.get("dns_servers") is not None:
        payload["properties"]["dhcpOptions"] = {
            "dnsServers": dhcp_options["dns_servers"]
        }
    return payload


def update_virtual_network_payload(
    hub,
    subscription_id: str,
    existing_payload: Dict[str, Any],
    new_values: Dict[str, Any],
) -> Dict[str, Any]:
    """
    Giving an existing resource state and desired state inputs, generate an updated payload, which can be used by
     PUT operation to update a resource on Azure.

    Args:
        hub: The redistributed pop central hub.
        subscription_id: subscription id required to update network security group for VN subnet
        existing_payload: An existing resource state from Azure. This is usually a GET operation response.
        new_values: A dictionary of desired state values. If any property's value is None,
         this property will be ignored. This is to match the behavior when a present() input is a None, Idem does not
         do an update.

    Returns:
        A result dict.
        result: True if no error occurs during the operation.
        ret: An updated payload that can be used to call PUT operation to update the resource. None if no update on all values.
        comment: A messages list.
    """
    result = {"result": True, "ret": None, "comment": []}
    is_updated = False
    new_payload = copy.deepcopy(existing_payload)
    if (new_values.get("tags") is not None) and (
        existing_payload.get("tags") != new_values.get("tags")
    ):
        new_payload["tags"] = new_values["tags"]
        is_updated = True
    if (new_values.get("extended_location") is not None) and (
        existing_payload.get("extendedLocation") != new_values.get("extended_location")
    ):
        new_payload["extendedLocation"] = new_values["extended_location"]
        is_updated = True
    existing_properties = existing_payload["properties"]
    if (new_values.get("address_space") is not None) and (
        set(new_values["address_space"])
        != set(existing_properties.get("addressSpace", {}).get("addressPrefixes", []))
    ):
        new_payload["properties"]["addressSpace"] = {
            "addressPrefixes": new_values["address_space"]
        }
        is_updated = True
    if (
        new_values.get("bgp_communities") is not None
        and new_values["bgp_communities"].get("virtual_network_community") is not None
    ) and (
        new_values["bgp_communities"]["virtual_network_community"]
        != existing_properties.get("bgpCommunities", {}).get("virtualNetworkCommunity")
    ):
        new_payload["properties"]["bgpCommunities"] = {
            "virtualNetworkCommunity": new_values["bgp_communities"].get(
                "virtual_network_community"
            )
        }
        is_updated = True
    if (new_values.get("flow_timeout_in_minutes") is not None) and (
        new_values["flow_timeout_in_minutes"]
        != existing_properties.get("flowTimeoutInMinutes")
    ):
        new_payload["properties"]["flowTimeoutInMinutes"] = new_values.get(
            "flow_timeout_in_minutes"
        )
        is_updated = True
    if (new_values.get("ddos_protection_plan") is not None) and (
        new_values["ddos_protection_plan"]
        != existing_properties.get("ddosProtectionPlan")
    ):
        new_payload["properties"]["ddosProtectionPlan"] = new_values.get(
            "ddos_protection_plan"
        )
        is_updated = True
    if (new_values.get("enable_ddos_protection") is not None) and (
        new_values["enable_ddos_protection"]
        != existing_properties.get("enableDdosProtection")
    ):
        new_payload["properties"]["enableDdosProtection"] = new_values.get(
            "enable_ddos_protection"
        )
        is_updated = True
    if (new_values.get("enable_vm_protection") is not None) and (
        new_values["enable_vm_protection"]
        != existing_properties.get("enableVmProtection")
    ):
        new_payload["properties"]["enableVmProtection"] = new_values.get(
            "enable_vm_protection"
        )
        is_updated = True
    if new_values.get("subnets") is not None:
        existing_subnets = existing_properties.get("subnets")
        existing_subnets_required_payload = convert_raw_to_present_vn_subnet(
            subnets=existing_subnets
        )
        if diff_virtual_network_subnets(
            new_values.get("subnets"), existing_subnets_required_payload, "name"
        ):
            new_payload["properties"]["subnets"] = convert_present_to_raw_vn_subnet(
                subscription_id=subscription_id, subnets=new_values.get("subnets")
            )
            is_updated = True
    if (
        new_values.get("dhcp_options") is not None
        and new_values.get("dhcp_options", {}).get("dns_servers") is not None
        and new_values.get("dhcp_options", {}).get("dns_servers")
        != existing_properties.get("dhcpOptions", {}).get("dnsServers")
    ):
        new_payload["properties"]["dhcpOptions"] = {
            "dnsServers": new_values.get("dhcp_options", {}).get("dns_servers")
        }
        is_updated = True
    if is_updated:
        result["ret"] = new_payload
    return result


def convert_present_to_raw_vn_subnet(
    subscription_id: str, subnets: List[Dict[str, Any]]
):
    """
    Giving some present function inputs, generate a payload that can be used during PUT operation to Azure. Any None
    value input will be ignored, unless this parameter is a required input parameter.

    Args:
        subscription_id : subscription id to create network security group id
        subnets(list[dict], optional): List of Subnet in a virtual network resource.Each Subnet will have fields
        ((name(str), required), (address_space(str, required) and service_endpoints(list), optional))

    Returns:
        Subnets List(Dict[str,any]) in the format of an Azure PUT operation payload.
    """
    raw_subnets: List = []
    for subnet in subnets:
        payload = {
            "name": subnet["name"],
        }
        if subnet.get("address_prefix") is not None:
            payload.setdefault("properties", {})["addressPrefix"] = subnet[
                "address_prefix"
            ]
        if subnet.get("address_prefixes") is not None:
            payload.setdefault("properties", {})["addressPrefixes"] = subnet[
                "address_prefixes"
            ]
        if subnet.get("security_group_id") is not None:
            payload.setdefault("properties", {})["networkSecurityGroup"] = {
                "id": subnet["security_group_id"]
            }
        if subnet.get("service_endpoints") is not None:
            services: List = []
            for endpoint in subnet.get("service_endpoints"):
                services.append({"service": endpoint})
            payload.setdefault("properties", {})["serviceEndpoints"] = services
        raw_subnets.append(payload)
    return raw_subnets


def convert_raw_to_present_vn_subnet(subnets: List[Dict[str, Any]]):
    """
     Giving an existing resource state and desired state inputs, generate a dict that match the format of
     present input parameters.

    Args:
        subnets(list[dict], optional): Resource List of Subnet in a virtual network resource.

    Returns:
         A subnet List that contains the parameters that match respective present function's input format.
    """
    present_subnets: List = []
    for subnet in subnets:
        new_subnet_payload = {"name": subnet.get("name")}
        if subnet["properties"].get("addressPrefix") is not None:
            new_subnet_payload["address_prefix"] = subnet["properties"]["addressPrefix"]
        if subnet["properties"].get("addressPrefixes") is not None:
            new_subnet_payload["address_prefixes"] = subnet["properties"][
                "addressPrefixes"
            ]
        if subnet["properties"].get("networkSecurityGroup") is not None:
            new_subnet_payload["security_group_id"] = subnet["properties"][
                "networkSecurityGroup"
            ].get("id")
        if subnet["properties"].get("serviceEndpoints") is not None:
            services: List = []
            for endpoint in subnet["properties"]["serviceEndpoints"]:
                services.append(endpoint["service"])
            new_subnet_payload["service_endpoints"] = services
        present_subnets.append(new_subnet_payload)
    return present_subnets


def diff_virtual_network_subnets(
    new_values: List[Dict[str, any]],
    resource: List[Dict[str, any]],
    sorting_key: str,
):
    """
    Compares virtual network subnets to check whether any of the state attributes has been added or modified.
    Returns true if there is any updates else false.

    Args:
        hub: The redistributed pop central hub.
        new_values: (List[Dict[str, any]]) Present value which will be given as input
        resource: (List[Dict[str, any]]) Raw resource response which needs to be compared with new_values
        sorting_key: (str) Primary/Unique key name within each dictionary , which will be used to sort dictionary
         objects with given list before comparing.

    Returns:
        A boolean value, True if there is any difference between List[Dict] arguments else returns False
    """
    for entry in new_values:
        for key, value in entry.items():
            if isinstance(entry[key], list):
                entry[key].sort()
    return sorted(new_values, key=lambda x: x[sorting_key]) != sorted(
        resource, key=lambda x: x[sorting_key]
    )



"""Exec module for managing Compute Disks."""

__func_alias__ = {"list_": "list"}

from collections import OrderedDict
from typing import Dict, Any


async def get(
    hub, ctx, resource_id: str, name: str = None, raw: bool = False
) -> Dict[str, Any]:
    """Get compute disks resource by resource_id.

    Args:
        resource_id(str):
            The resource_id of a disk
        name(str, Optional):
            The name of the resource
        raw(bool, Optional):
            Returns raw response if True. Defaults to False

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.compute.disks.get resource_id="value"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path:  azure.compute.disks.get
                - kwargs:
                    resource_id: "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Compute/disks/{disk_name}"

    """
    result = dict(comment=[], result=True, ret=None)
    uri_parameters = OrderedDict(
        {
            "subscriptions": "subscriptionId",
            "resourceGroups": "resourceGroupName",
            "disks": "diskName",
        }
    )

    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{hub.exec.azure.URL}{resource_id}?api-version={hub.tool.azure.api_versions.compute.disks}",
        success_codes=[200],
    )
    if not response_get["result"]:
        if response_get["status"] != 404:
            result["result"] = False
        result["comment"].extend(
            hub.tool.azure.result_utils.extract_error_comments(response_get)
        )
        return result

    elif response_get["result"] and response_get["ret"]:
        if raw:
            result["ret"] = response_get["ret"]
        else:
            uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
                resource_id, uri_parameters
            )
            result["ret"] = hub.tool.azure.compute.disks.convert_raw_to_present_state(
                {**uri_parameter_values, **response_get["ret"]}
            )

            result["ret"]["resource_id"] = resource_id

    return result


async def list_(hub, ctx) -> Dict:
    """List of compute disks.

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.compute.disks.list

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: azure.compute.disks.list


    """
    result = dict(comment=[], result=True, ret=[])
    subscription_id = ctx.acct.subscription_id
    uri_parameters = OrderedDict(
        {
            "subscriptions": "subscriptionId",
            "resourceGroups": "resourceGroupName",
            "disks": "diskName",
        }
    )
    async for page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}"
        f"/providers/Microsoft.Compute/disks?api-version={hub.tool.azure.api_versions.compute.disks}",
        success_codes=[200],
    ):
        resource_list = page_result.get("value")
        if resource_list:
            for resource_raw_state in resource_list:
                resource_id = resource_raw_state["id"]
                uri_parameter_values = hub.tool.azure.uri.get_parameter_value_in_dict(
                    resource_id, uri_parameters
                )
                resource_present_state = (
                    hub.tool.azure.compute.disks.convert_raw_to_present_state(
                        {**uri_parameter_values, **resource_raw_state}
                    )
                )
                resource_present_state["resource_id"] = resource_id
                resource_present_state["name"] = resource_id
                result["ret"].append(resource_present_state)

    return result



"""Exec module for managing network security_rules."""
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub, ctx, resource_id: str, name: str = None, raw: bool = False
) -> Dict[str, Any]:
    """Get network security rule resource from resource_id.

    Args:
        resource_id(str):
            The resource_id of security rule
        name(str, Optional):
            The name of the resource
        raw(bool, Optional):
            Returns raw response if True. Defaults to False

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.network.security_rules.get resource_id="value"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path:  azure.network.security_rules.get
                - kwargs:
                    resource_id: "/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/networkSecurityGroups/{network_security_group_name}/securityRules/{security_rule_name}"

    """
    result = dict(comment=[], result=True, ret=None)

    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2022-07-01",
        success_codes=[200],
    )
    if not response_get["result"]:
        if response_get["status"] != 404:
            result["result"] = False
        result["comment"].extend(
            hub.tool.azure.result_utils.extract_error_comments(response_get)
        )
        return result

    elif response_get["result"] and response_get["ret"]:
        if raw:
            result["ret"] = response_get["ret"]
        else:
            result[
                "ret"
            ] = hub.tool.azure.network.security_rules.convert_raw_security_rules_to_present(
                resource=response_get["ret"], idem_resource_name=resource_id
            )

    return result


async def list_(hub, ctx) -> Dict:
    """List of network security rules

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.network.security_rules.list

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: azure.network.security_rules.list


    """
    result = dict(comment=[], result=True, ret=[])
    subscription_id = ctx.acct.subscription_id
    async for nsg_page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/providers/Microsoft.Network/networkSecurityGroups?api-version=2021-03-01",
        success_codes=[200],
    ):
        nsg_resource_list = nsg_page_result.get("value")
        if nsg_resource_list:
            for nsg_resource in nsg_resource_list:
                if nsg_resource.get("properties") and nsg_resource.get(
                    "properties"
                ).get("securityRules"):
                    for resource in nsg_resource.get("properties").get("securityRules"):
                        resource_id = resource["id"]
                        result["ret"].append(
                            hub.tool.azure.network.security_rules.convert_raw_security_rules_to_present(
                                resource=resource,
                                idem_resource_name=resource_id,
                            )
                        )
    return result



PRESENT_TO_RAW_MAP = {
    "subscriptionId": "subscription_id",
    "resourceGroupName": "resource_group_name",
    "networkInterfaceName": "network_interface_name",
    "tags": "tags",
    "location": "location",
    "properties": {
        "ipConfigurations": [
            {
                "$list_refs": ["ip_configurations"],
                "name": "ip_configurations.name",
                "properties": {
                    "privateIPAllocationMethod": "ip_configurations.private_ip_address_allocation",
                    "subnet": {"id": "ip_configurations.subnet_id"},
                    "privateIPAddressVersion": "ip_configurations.private_ip_address_version",
                    "privateIPAddress": "ip_configurations.private_ip_address",
                    "publicIPAddress": {"id": "ip_configurations.public_ip_address_id"},
                    "primary": "ip_configurations.primary",
                },
            }
        ],
        "enableAcceleratedNetworking": "enable_accelerated_networking",
        "dnsSettings": {
            "dnsServers": "dns_settings.dns_servers",
            "internalDnsNameLabel": "dns_settings.internal_dns_name_label",
        },
        "networkSecurityGroup": {
            "id": "network_security_group_id",
        },
    },
}


RAW_TO_PRESENT_MAP = {
    "subscription_id": "subscriptionId",
    "resource_group_name": "resourceGroupName",
    "network_interface_name": "networkInterfaceName",
    "tags": "tags",
    "location": "location",
    "ip_configurations": [
        {
            "$list_refs": ["properties.ipConfigurations"],
            "name": "properties.ipConfigurations.name",
            "private_ip_address_allocation": "properties.ipConfigurations.properties.privateIPAllocationMethod",
            "subnet_id": "properties.ipConfigurations.properties.subnet.id",
            "private_ip_address_version": "properties.ipConfigurations.properties.privateIPAddressVersion",
            "private_ip_address": "properties.ipConfigurations.properties.privateIPAddress",
            "public_ip_address_id": "properties.ipConfigurations.properties.publicIPAddress.id",
            "primary": "properties.ipConfigurations.properties.primary",
        }
    ],
    "enable_accelerated_networking": "properties.enableAcceleratedNetworking",
    "dns_settings": {
        "dns_servers": "properties.dnsSettings.dnsServers",
        "internal_dns_name_label": "properties.dnsSettings.internalDnsNameLabel",
    },
    "network_security_group_id": "properties.networkSecurityGroup.id",
}


def convert_raw_to_present_state(hub, raw_state):
    return hub.tool.azure.generic.convert_state_format(
        raw_state,
        RAW_TO_PRESENT_MAP,
    )


def convert_present_to_raw_state(hub, present_state):
    return hub.tool.azure.generic.convert_state_format(
        present_state,
        PRESENT_TO_RAW_MAP,
    )



"""
Autogenerated using `pop-create-idem <https://gitlab.com/saltstack/pop/pop-create-idem>`__


"""
from collections import OrderedDict
from typing import Any
from typing import Dict

__contracts__ = ["resource"]


async def present(
    hub,
    ctx,
    name: str,
    resource_group_name: str,
    public_ip_address_name: str,
    parameters: dict = None,
    force_update: bool = False,
):
    r"""
    **Autogenerated function**

    Create or update Public IP Addresses

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        public_ip_address_name(str): The name of the public IP address.
        parameters(dict, optional): API request payload parameters. Defaults to {}.
        force_update(bool, optional): If PUT operation should be used instead of PATCH operation during resource update. Defaults to False.

    Returns:
        dict

    Examples:

        .. code-block:: sls

            resource_is_present:
              azure.virtual_networks.public_ip_addresses.present:
                - name: value
                - resource_group_name: value
                - public_ip_address_name: value
    """
    if parameters is None:
        parameters = {}

    subscription_id = ctx.acct.subscription_id
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/publicIPAddresses/{public_ip_address_name}?api-version=2021-03-01",
        success_codes=[200],
    )

    if force_update:
        if ctx.get("test", False):
            return dict(
                name=name,
                result=True,
                comment="Would force to update azure.virtual_networks.public_ip_addresses",
            )
        response_force_put = await hub.exec.request.json.put(
            ctx,
            url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/publicIPAddresses/{public_ip_address_name}?api-version=2021-03-01",
            success_codes=[200, 201],
            json=parameters,
        )
        if response_force_put["result"]:
            old_resource = response_get["ret"] if response_get["result"] else None
            return dict(
                name=name,
                result=True,
                old_state=old_resource,
                new_state=response_force_put["ret"],
                comment=response_force_put["comment"],
            )
        else:
            hub.log.debug(
                f"Could not force to update Public IP Addresses {response_force_put['comment']} {response_force_put['ret']}"
            )
            return dict(
                name=name,
                result=False,
                comment=response_force_put["comment"],
                error=response_force_put["ret"],
            )

    if not response_get["result"]:
        if ctx.get("test", False):
            return dict(
                name=name,
                result=True,
                comment="Would create azure.virtual_networks.public_ip_addresses",
            )

        if response_get["status"] == 404:
            # PUT operation to create a resource
            response_put = await hub.exec.request.json.put(
                ctx,
                url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/publicIPAddresses/{public_ip_address_name}?api-version=2021-03-01",
                success_codes=[200, 201],
                json=parameters,
            )

            if not response_put["result"]:
                hub.log.debug(
                    f"Could not create Public IP Addresses {response_put['comment']} {response_put['ret']}"
                )
                return dict(
                    name=name,
                    result=False,
                    comment=response_put["comment"],
                    error=response_put["ret"],
                )

            return dict(
                name=name,
                result=True,
                old_state=None,
                new_state=response_put["ret"],
                comment=response_put["comment"],
            )
        else:
            hub.log.debug(
                f"Could not get Public IP Addresses {response_get['comment']} {response_get['ret']}"
            )
            return dict(
                name=name,
                result=False,
                comment=response_get["comment"],
                error=response_get["ret"],
            )
    else:
        # PATCH operation to update a resource
        patch_parameters = {"tags": "tags"}
        existing_resource = response_get["ret"]
        new_parameters = hub.tool.azure.request.patch_json_content(
            patch_parameters, existing_resource, parameters
        )
        if ctx.get("test", False):
            return dict(
                name=name,
                result=True,
                comment=f"Would update azure.virtual_networks.public_ip_addresses with parameters: {new_parameters}",
            )

        if not new_parameters:
            return dict(
                name=name,
                result=True,
                old_state=existing_resource,
                new_state=existing_resource,
                comment=f"'{name}' has no property need to be updated.",
            )

        response_patch = await hub.exec.request.json.patch(
            ctx,
            url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/publicIPAddresses/{public_ip_address_name}?api-version=2021-03-01",
            success_codes=[200],
            json=new_parameters,
        )

        if not response_patch["result"]:
            hub.log.debug(
                f"Could not update Public IP Addresses {response_patch['comment']} {response_patch['ret']}"
            )
            return dict(
                name=name,
                result=False,
                comment=response_patch["comment"],
                error=response_patch["ret"],
            )

        return dict(
            name=name,
            result=True,
            old_state=existing_resource,
            new_state=response_patch["ret"],
            comment=response_patch["comment"],
        )


async def absent(
    hub, ctx, name: str, resource_group_name: str, public_ip_address_name: str
) -> dict:
    r"""
    **Autogenerated function**

    Delete Public IP Addresses

    Args:
        name(str): The identifier for this state.
        resource_group_name(str): The name of the resource group.
        public_ip_address_name(str): The name of the public IP address.

    Returns:
        dict

    Examples:

        .. code-block:: sls

            resource_is_absent:
              azure.virtual_networks.public_ip_addresses.absent:
                - name: value
                - resource_group_name: value
                - public_ip_address_name: value
    """

    subscription_id = ctx.acct.subscription_id
    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/publicIPAddresses/{public_ip_address_name}?api-version=2021-03-01",
        success_codes=[200],
    )
    if response_get["result"]:
        if ctx.get("test", False):
            return dict(
                name=name,
                result=True,
                comment="Would delete azure.virtual_networks.public_ip_addresses",
            )

        existing_resource = response_get["ret"]
        response_delete = await hub.exec.request.raw.delete(
            ctx,
            url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/resourceGroups/{resource_group_name}/providers/Microsoft.Network/publicIPAddresses/{public_ip_address_name}?api-version=2021-03-01",
            success_codes=[200, 202, 204],
        )

        if not response_delete["result"]:
            hub.log.debug(
                f"Could not delete Public IP Addresses {response_delete['comment']} {response_delete['ret']}"
            )
            return dict(
                name=name,
                result=False,
                comment=response_delete["comment"],
                error=response_delete["ret"],
            )

        return dict(
            name=name,
            result=True,
            old_state=existing_resource,
            new_state={},
            comment=response_delete["comment"],
        )
    elif response_get["status"] == 404:
        # If Azure returns 'Not Found' error, it means the resource has been absent.
        return dict(
            name=name,
            result=True,
            old_state=None,
            new_state=None,
            comment=f"'{name}' already absent",
        )
    else:
        hub.log.debug(
            f"Could not get Public IP Addresses {response_get['comment']} {response_get['ret']}"
        )
        return dict(
            name=name,
            result=False,
            comment=response_get["comment"],
            error=response_get["ret"],
        )


async def describe(hub, ctx) -> Dict[str, Dict[str, Any]]:
    r"""
    **Autogenerated function**

    Describe the resource in a way that can be recreated/managed with the corresponding "present" function


    List all Public IP Addresses under the same subscription


    Returns:
        Dict[str, Any]

    Examples:

        .. code-block:: bash

            $ idem describe azure_auto.virtual_networks.public_ip_addresses
    """

    result = {}
    subscription_id = ctx.acct.subscription_id
    uri_parameters = OrderedDict(
        {
            "resourceGroups": "resource_group_name",
            "publicIPAddresses": "public_ip_address_name",
        }
    )
    async for page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}/providers/Microsoft.Network/publicIPAddresses?api-version=2021-03-01",
        success_codes=[200],
    ):
        resource_list = page_result.get("value", None)
        if resource_list:
            for resource in resource_list:
                uri_parameter_values = hub.tool.azure.uri.get_parameter_value(
                    resource["id"], uri_parameters
                )
                result[resource["id"]] = {
                    f"azure.virtual_networks.public_ip_addresses.present": uri_parameter_values
                    + [{"parameters": resource}]
                }
    return result



"""Exec module for managing policy assignments."""
from typing import Any
from typing import Dict

__func_alias__ = {"list_": "list"}


async def get(
    hub,
    ctx,
    resource_id: str,
    name: str = None,
    raw: bool = False,
) -> Dict[str, Any]:
    """Get policy assignments resource from resource_id.

    Args:
        resource_id(str):
            The resource_id of policy assignments
        name(str, Optional):
            The name of the resource
        raw(bool, Optional):
            Returns raw response if True. Defaults to False

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.policy.policy_assignments.get resource_id="value" raw="False"

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path:  azure.policy.policy_assignments.get
                - kwargs:
                    resource_id: "{scope}/providers/Microsoft.Authorization/policyAssignments/{policy_assignment_name}"
                    raw: False

    """
    result = dict(comment=[], result=True, ret=None)

    response_get = await hub.exec.request.json.get(
        ctx,
        url=f"{ctx.acct.endpoint_url}{resource_id}?api-version=2021-06-01",
        success_codes=[200],
    )
    if not response_get["result"]:
        if response_get["status"] != 404:
            result["result"] = False
        result["comment"].extend(
            hub.tool.azure.result_utils.extract_error_comments(response_get)
        )
        return result

    if raw:
        result["ret"] = response_get["ret"]

    else:
        result[
            "ret"
        ] = hub.tool.azure.policy.policy_assignment.convert_raw_policy_assignment_to_present(
            resource=response_get["ret"],
            idem_resource_name=resource_id,
            resource_id=resource_id,
        )

    return result


async def list_(hub, ctx) -> Dict:
    """List of policy assignments

    Returns:
        Dict[str, Any]

    Examples:
        Calling this exec module function from the cli with resource_id:

        .. code-block:: bash

            idem exec azure.policy.policy_assignments.list

        Using in a state:

        .. code-block:: yaml

            my_unmanaged_resource:
              exec.run:
                - path: azure.policy.policy_assignments1.list


    """
    result = dict(comment=[], result=True, ret=[])
    subscription_id = ctx.acct.subscription_id
    async for page_result in hub.tool.azure.request.paginate(
        ctx,
        url=f"{ctx.acct.endpoint_url}/subscriptions/{subscription_id}"
        f"/providers/Microsoft.Authorization/policyAssignments?api-version=2021-06-01",
        success_codes=[200],
    ):
        resource_list = page_result.get("value")
        if resource_list:
            for resource in resource_list:
                resource_id = resource["id"]
                result["ret"].append(
                    hub.tool.azure.policy.policy_assignment.convert_raw_policy_assignment_to_present(
                        resource=resource,
                        idem_resource_name=resource_id,
                        resource_id=resource_id,
                    )
                )
    return result



from dict_tools.data import NamespaceDict

# Azure API versions by service/resource

# In plugin code, API version can be accessed either as a constant/expression or as a function call result.
# Sample code (both variants should return api_version "2021-12-01"):
#   api_version = hub.tool.azure.api_versions.compute.disks
#   api_version = hub.tool.azure.api_versions.get_api_version('compute.disks')


def get_api_version(hub, resource_type: str) -> str:
    if resource_type.startswith("azure."):
        resource_type = resource_type.replace("azure.", "")

    ref = hub.tool.azure.api_versions
    for path_segment in resource_type.split("."):
        ref = ref[path_segment]

    return ref


compute = NamespaceDict(
    {
        "disks": "2021-12-01",
        "virtual_machines": "2022-03-01",
    }
)

sql_database = NamespaceDict(
    {
        "databases": "2021-11-01",
    }
)

network = NamespaceDict(
    {
        "network_interfaces": "2021-08-01",
        "virtual_networks": "2022-07-01",
        "virtual_network_peerings": "2021-03-01",
    }
)
